#!/usr/bin/env python

import nsysstats

class CUDAGPUStarvation(nsysstats.Report):

    ROW_LIMIT = 50
    GAP = 500

    usage = f"""{{SCRIPT}}[:rows=<limit>][:start=<time>][:end=<time>][:nvtx=<range[@domain]>][:gap=<threshold>] -- CUDA GPU Starvation

    Options:
        rows=<limit> - Maximum number of rows returned by the query.
            Default is {ROW_LIMIT}.

        start=<time> - Start time used for filtering in nanoseconds.

        end=<time> - End time used for filtering in nanoseconds.

        nvtx=<range[@domain]> - NVTX range text and domain used for filtering.
            Do not specify the domain for ranges in the default domain. Note
            that only the first matching record will be considered.
            If this option is used along with the 'start' and/or 'end' options,
            the explicit start/end times will override the NVTX range times.

        gap=<threshold> - Minimum duration of GPU gaps in milliseconds.
            Default is {GAP}ms.

    Output: All time values default to nanoseconds
        Row# : Row number of GPU gap
        Duration : Duration of GPU gap
        Start : Start time of GPU gap
        PID : Process identifier
        Device ID : GPU device identifier

    This rule identifies time regions where a GPU is idle for longer than a set
    threshold. For each process, each GPU device is examined, and gaps are
    found within the time range that starts with the beginning of the first GPU
    operation on that device and ends with the end of the last GPU operation on
    that device.

    GPU gaps that cannot be addressed by the user are excluded. This includes:
        1. CUDA profiling overhead in the middle of a GPU gap.
        2. The initial gap starting before the first GPU operation, which
           corresponds to the CUDA initialization overhead.
        3. The final gap starting after the last GPU operation, which
           corresponds to the flushing of CUPTI buffers when stopping the
           collection.
"""

    query_format_columns = """
    SELECT
        ROW_NUMBER() OVER(ORDER BY duration DESC, gapStart) AS "Row#",
        duration AS "Duration:dur_ns",
        gapStart AS "Start:ts_ns",
        pid AS "PID",
        deviceId AS "Device ID",
        globalPid AS "_Global ID"
    FROM
        ({GAPS})
"""

# Find gaps.
# "ops" is the table containing kernel + memory + CUDA profiling overhead.
# 1. CTE "starts": Give a rowNum, SRi, to each start, ordered by start time.
# 2. CTE "ends": Give a rowNum, ERj, to each end, ordered by end time.
# 3. Reconstruct intervals [ERj, SRj+1] by putting together an end ERj with the
#    next start SRj+1 (start_rowNum - 1 = end_rowNum).
# 4. Keep only those intervals [ERj, SRj+1] that are valid (ERj < SRj+1).
#
# Assume that we have the following intervals:
#
# SR1                          ER2
#  |--------------a-------------|
#      SR2                ER1
#       |---------b--------|
#                                         SR3              ER3
#                                          |--------c-------|
# With step 3, we get:
# 1. ER1 joined with SR2.
# 2. ER2 joined with SR3.
#
#      SR2                 ER1
#       |---------a'--------|
#                               ER2        SR3
#                                |----b'----|
#
# Only the second interval (b') meets the condition end < start of step 4 and
# will be considered as a gap. (a') will be discarded and the query will
# return:
#
#                               ER2        SR3
#                                |----b'----|
#
# ER2 will be the start and SR3 will be the end of the gap.
    query_gap = """
    WITH
        ops AS (
            {OPS_ALL}
        ),
        starts AS (
            SELECT
                ROW_NUMBER() OVER(ORDER BY pid, deviceId, start) AS rowNum,
                start,
                pid,
                globalPid,
                deviceId
            FROM
                ops
        ),
        ends AS (
            SELECT
                ROW_NUMBER() OVER(ORDER BY pid, deviceId, end) AS rowNum,
                end,
                pid,
                globalPid,
                deviceId
            FROM
                ops
        )
    SELECT
        start - end AS duration,
        end AS gapStart,
        start AS gapEnd,
        starts.pid,
        starts.globalPid,
        starts.deviceId
    FROM
        starts
    JOIN
        ends
        ON      starts.rowNum - 1 == ends.rowNum
            AND starts.deviceId == ends.deviceId
            AND starts.pid == ends.pid
    WHERE
            duration > {THRESHOLD}
        AND gapStart < gapEnd
    LIMIT {ROW_LIMIT}
"""

# Select columns of kernel/memory operations.
    query_select = """
    SELECT
        start,
        end,
        (globalPid >> 24) & 0x00FFFFFF AS pid,
        globalPid,
        deviceId
    FROM
        {GPU_OPERATION}
"""

# Combine kernel/memory operations.
    query_union = """
        UNION ALL
"""

# Add the profiling overhead to the GPU operation table returned by
# "query_union".
# 1. CTE "range": Get [min(start), max(end)] for each deviceId/PID. It will be
#    used as the clipping range for overheads.
# 2. CTE "overhead": Select the profiling overhead that we want to take into
#    account.
# 3. Duplicate overhead rows for each deviceId/PID. This will create a deviceId
#    column that is not initially in the PROFILER_OVERHEAD table.
#    Note: a profiling overhead on one thread affects all GPUs of the same
#    process.
# 4. The overhead rows are combined with GPU operation rows.
    query_overhead = """
    WITH
        gpuops AS (
            {GPU_OPS}
        ),
        range AS (
            SELECT
                min(start) AS start,
                max(end) AS end,
                pid,
                globalPid,
                deviceId
            FROM
                gpuops
            GROUP BY deviceId, pid
        ),
        overheadID AS (
            SELECT
                id
            FROM
                StringIds
            WHERE
                   value == 'CUDA profiling data flush overhead'
                OR value == 'CUDA profiling stop overhead'
                OR value == 'CUDA profiling overhead'
        ),
        overhead AS (
            SELECT
                po.start,
                po.end,
                (po.globalTid >> 24) & 0x00FFFFFF AS pid
            FROM
                PROFILER_OVERHEAD AS po
            JOIN
                overheadID AS co
                ON co.id == po.nameId
        )
    SELECT
        co.start,
        co.end,
        co.pid,
        range.globalPid,
        range.deviceId
    FROM
        overhead AS co
    JOIN
        range
        ON      co.pid == range.pid
            AND co.start > range.start
            AND co.end < range.end
    UNION ALL
    SELECT
        *
    FROM
        gpuops
"""

    def setup(self):
        err = super().setup()
        if err != None:
            return err

        parser = self.ArgumentParser(allow_abbrev=False)
        parser.add_argument('--rows', default=self.ROW_LIMIT, type=int, \
            help='maximum number of rows returned by the query')
        parser.add_argument('--start', type=int, \
            help='start time used for filtering')
        parser.add_argument('--end', type=int, \
            help='end time used for filtering')
        parser.add_argument('--nvtx', type=str, \
            help='NVTX range and domain used for filtering')
        parser.add_argument('--gap', default=self.GAP, type=int, \
            help='maximum percentage of time the GPU is being used')
        args = parser.parse_expertsys_options(self.args, {'gap'})

        err = self.filter_time_range(args.start, args.end, args.nvtx)
        if err != None:
            return err

        sub_queries = []

        kernel = 'CUPTI_ACTIVITY_KIND_KERNEL'
        memcpy = 'CUPTI_ACTIVITY_KIND_MEMCPY'
        memset = 'CUPTI_ACTIVITY_KIND_MEMSET'

        if self.table_exists(kernel):
            sub_queries.append(self.query_select.format(GPU_OPERATION = kernel))

        if self.table_exists(memcpy):
            sub_queries.append(self.query_select.format(GPU_OPERATION = memcpy))

        if self.table_exists(memset):
            sub_queries.append(self.query_select.format(GPU_OPERATION = memset))

        if len(sub_queries) == 0:
            return "{DBFILE} could not be analyzed because it does not contain the required CUDA data." \
                " Does the application launch CUDA GPU operations?"

        ops = self.query_union.join(sub_queries)

        if self.table_exists('PROFILER_OVERHEAD'):
            ops = self.query_overhead.format(
                GPU_OPS = ops)

        gaps = self.query_gap.format(
            OPS_ALL = ops,
            ROW_LIMIT = args.rows,
            # ms to ns
            THRESHOLD = args.gap * 1000000)

        self.query = self.query_format_columns.format(
            GAPS = gaps)

if __name__ == "__main__":
    CUDAGPUStarvation.Main()
