#!/usr/bin/env python

import nsysstats

class VULKANGPUStarvation(nsysstats.Report):

    ROW_LIMIT = 50
    GAP = 500

    usage = f"""{{SCRIPT}}[:rows=<limit>][:start=<time>][:end=<time>][:nvtx=<range[@domain]>][:gap=<threshold>] -- VULKAN GPU Starvation

    Options:
        rows=<limit> - Maximum number of rows returned by the query.
            Default is {ROW_LIMIT}.

        start=<time> - Start time used for filtering in nanoseconds.

        end=<time> - End time used for filtering in nanoseconds.

        nvtx=<range[@domain]> - NVTX range text and domain used for filtering.
            Do not specify the domain for ranges in the default domain. Note
            that only the first matching record will be considered.
            If this option is used along with the 'start' and/or 'end' options,
            the explicit start/end times will override the NVTX range times.

        gap=<threshold> - Minimum duration of GPU gaps in milliseconds.
            Default is {GAP}ms.

    Output: All time values default to nanoseconds
        Row# : Row number of GPU gap
        Duration : Duration of GPU gap
        Start : Start time of GPU gap
        PID : Process identifier
        Device ID : GPU device identifier

    This rule identifies time regions where a GPU is idle for longer than a set
    threshold. For each process, each GPU device is examined, and gaps are
    found within the time range that starts with the beginning of the first GPU
    command on that device and ends with the end of the last GPU command on
    that device. VULKAN profiling overheads are excluded from the gaps.
"""

    query_format_columns = """
    SELECT
        ROW_NUMBER() OVER(ORDER BY duration DESC, gapStart) AS "Row#",
        duration AS "Duration:dur_ns",
        gapStart AS "Start:ts_ns",
        pid AS "PID",
        deviceId AS "Device ID",
        globalTid AS "_Global ID"
    FROM
        ({GAPS})
"""

    query_gap = """
    WITH
        ops AS (
            {OPS_ALL}
        ),
        starts AS (
            SELECT
                ROW_NUMBER() OVER(ORDER BY pid, deviceId, start) AS rowNum,
                start,
                pid,
                globalTid,
                deviceId
            FROM
                ops
        ),
        ends AS (
            SELECT
                ROW_NUMBER() OVER(ORDER BY pid, deviceId, end) AS rowNum,
                end,
                pid,
                globalTid,
                deviceId
            FROM
                ops
        )
    SELECT
        start - end AS duration,
        end AS gapStart,
        start AS gapEnd,
        starts.pid,
        starts.globalTid,
        starts.deviceId
    FROM
        starts
    JOIN
        ends
        ON      starts.rowNum - 1 == ends.rowNum
            AND starts.deviceId == ends.deviceId
            AND starts.pid == ends.pid
    WHERE
            duration > {THRESHOLD}
        AND gapStart < gapEnd
    LIMIT {ROW_LIMIT}
"""

    query_select = """
    SELECT
        start,
        end,
        (globalTid >> 24) & 0x00FFFFFF AS pid,
        globalTid,
        deviceId
    FROM
        VULKAN_WORKLOAD
"""

    query_overhead = """
    WITH
        gpuops AS (
            {GPU_OPS}
        ),
        range AS (
            SELECT
                min(start) AS start,
                max(end) AS end,
                pid,
                globalTid,
                deviceId
            FROM
                gpuops
            GROUP BY deviceId, pid
        ),
        overheadID AS (
            SELECT
                id
            FROM
                StringIds
            WHERE
                value == 'Vulkan profiling overhead'
        ),
        overhead AS (
            SELECT
                po.start,
                po.end,
                (po.globalTid >> 24) & 0x00FFFFFF AS pid
            FROM
                PROFILER_OVERHEAD AS po
            JOIN
                overheadID AS co
                ON co.id == po.nameId
        )
    SELECT
        co.start,
        co.end,
        co.pid,
        range.globalTid,
        range.deviceId
    FROM
        overhead AS co
    JOIN
        range
        ON      co.pid == range.pid
            AND co.start > range.start
            AND co.end < range.end
    UNION ALL
    SELECT
        *
    FROM
        gpuops
"""

    table_checks = {
        'VULKAN_WORKLOAD':
            "{DBFILE} could not be analyzed because it does not contain the required VULKAN data."
            " Does the application submit commands to the GPU?"
    }

    table_col_checks = {
        'VULKAN_WORKLOAD':
        { 'deviceId':
            "{DBFILE} could not be analyzed due to missing deviceId."
            " Please re-export the report file with a recent version of Nsight Systems." }
    }

    def setup(self):
        err = super().setup()
        if err != None:
            return err

        parser = self.ArgumentParser(allow_abbrev=False)
        parser.add_argument('--rows', default=self.ROW_LIMIT, type=int, \
            help='maximum number of rows returned by the query')
        parser.add_argument('--start', type=int, \
            help='start time used for filtering')
        parser.add_argument('--end', type=int, \
            help='end time used for filtering')
        parser.add_argument('--nvtx', type=str, \
            help='NVTX range and domain used for filtering')
        parser.add_argument('--gap', default=self.GAP, type=int, \
            help='maximum percentage of time the GPU is being used')
        args = parser.parse_expertsys_options(self.args, {'gap'})

        err = self.filter_time_range(args.start, args.end, args.nvtx)
        if err != None:
            return err

        ops = self.query_select

        if self.table_exists('PROFILER_OVERHEAD'):
            ops = self.query_overhead.format(
                GPU_OPS = ops)

        gaps = self.query_gap.format(
            OPS_ALL = ops,
            ROW_LIMIT = args.rows,
            # ms to ns
            THRESHOLD = args.gap * 1000000)

        self.query = self.query_format_columns.format(
            GAPS = gaps)

if __name__ == "__main__":
    VULKANGPUStarvation.Main()
